from .agent import agent, BaseAgent, Agent, returns_args, yields_args
from .team import AgentsTeam, team
from .routes import routes
from .goal_seaker import GoalSeakerTeam, goal_seaker_team
from ..visibility import pub, pvt


from vespwood import (
    MaxTokenLimitError, 
    MissingHookError, 
    MissingParamError, 
    MissingSchemaError, 
    MissingToolError, 
    MissingValidatorError, 
    PauseGeneration, 
    RateLimitError, 
    StopGeneration, 
    ValidationError,
)

__all__ = [
    "MaxTokenLimitError",
    "MissingHookError",
    "MissingParamError",
    "MissingSchemaError",
    "MissingToolError",
    "MissingValidatorError",
    "PauseGeneration",
    "RateLimitError",
    "StopGeneration",
    "ValidationError"
]


from typing import Any, Callable, List, Optional, get_type_hints

from vesp.invokation import Invokation
from .agent import BaseAgent
from .team import HandoverResponse, TeamLike, AgentsTeam


def routes(entrypoint: str = "/"):
    def decorator(func: Callable[[], TeamLike]):
        # Assert correct function type
        hints = get_type_hints(func)
        return_type = hints.get('return', None)
        if return_type is not TeamLike:
            raise TypeError("decorated function should return a TeamLike dict")
        
        # Actual decorator implementation
        class TeamWrapper(AgentsTeam):
            def __init__(self, *args, **kwargs):
                self.__name__ = func.__name__
                self.__doc__ = func.__doc__
                super().__init__(self.create_routes(), *args, **kwargs)


            @property
            def index(self) -> Optional[BaseAgent]:
                return self.get(entrypoint)
            
            def create_routes(self) -> TeamLike:
                return func()
            
            
            def handover(self, route: str, output: Any, chain: list[Invokation]) -> Optional[List[HandoverResponse]]:
                return None
            
        return TeamWrapper
    return decorator


from .agents import (
    BaseAgent, 
    Agent, 
    agent, 
    returns_args,
    yields_args,
    AgentsTeam,
    team
)
from .invokation import Invokation
from .visibility import Visibility

from vespwood import (
    Block, File, Image, Structured, ToolCall,
    Message, Prompt, Response, Tag, TaggedMessages,
    GeneratorClass, Generator,
    Schematic, schema, Schema, tool, Tool, hook, Hook, interceptor, Interceptor, validator, Validator,
    FormatObject, FormatList, FormatKeys,
    HookObject, HooksList, Params, PreparedArgs, Role, Saves, SchemaObject, SchemaInfo, ToolObject, ToolsList, ValidatorsList
)

from .errors import (
    MaxTokenLimitError, 
    MissingHookError, 
    MissingParamError, 
    MissingSchemaError, 
    MissingToolError, 
    MissingValidatorError, 
    PauseGeneration, 
    RateLimitError, 
    StopGeneration, 
    ValidationError,
)


__all__ = [
    "BaseAgent",
    "Agent",
    "agent",
    "AgentsTeam",
    "team",
    "Invokation",
    "Visibility",
    "Block",
    "File",
    "Image",
    "Structured",
    "ToolCall",
    "Message",
    "Prompt",
    "Response",
    "Tag",
    "TaggedMessages",
    "GeneratorClass",
    "Generator",
    "Schematic",
    "schema",
    "Schema",
    "tool",
    "Tool",
    "FormatObject",
    "FormatList",
    "FormatKeys",
    "hook",
    "Hook",
    "interceptor",
    "Interceptor",
    "validator",
    "Validator",
    "HookObject",
    "HooksList",
    "Params",
    "PreparedArgs",
    "Role",
    "Saves",
    "SchemaObject",
    "SchemaInfo",
    "ToolObject",
    "ToolsList",

    "ValidatorsList",
    "MaxTokenLimitError",
    "MissingHookError",
    "MissingParamError",
    "MissingSchemaError",
    "MissingToolError",
    "MissingValidatorError",
    "PauseGeneration",
    "RateLimitError",
    "StopGeneration",
    "ValidationError",
]


from __future__ import annotations
import asyncio
from typing import Callable
import uuid
from weakref import ReferenceType, ref


class AliveCountRef:
    __slots__ = '_count', '_dropped_to_zero', '_on_zero_alive_callbacks'

    def __init__(self, start: int = 0):
        self._count: int = start
        self._dropped_to_zero: bool = False
        self._on_zero_alive_callbacks: list[Callable[[], None]] = []


    @property
    def count(self) -> int:
        return self._count


    @property
    def dropped_to_zero(self) -> bool:
        return self._dropped_to_zero


    def increment(self):
        if not self._dropped_to_zero:
            self._count += 1


    def decrement(self):
        if not self.dropped_to_zero:
            if self._count == 0:
                raise ValueError("Alive count already dropped to 0")
            self._count -= 1
            if self._count == 0:
                self._dropped_to_zero = True
                for callback in self._on_zero_alive_callbacks: callback()


    def zero(self):
        self._count = 0
        self._dropped_to_zero = True
        for callback in self._on_zero_alive_callbacks: callback()


    def on_zero_alive(self, func: Callable[[], None]):
        self._on_zero_alive_callbacks.append(func)



class Output[T]:
    __slots__ = '_id', '_data', '_invokation_ref'
    def __init__(self, output: T, invokation: "Invokation"):
        self._id = uuid.uuid4().hex
        self._data: T = output
        self._invokation_ref: ReferenceType["Invokation"] = ref(invokation)

    @property
    def id(self) -> str:
        return self._id
    
    @property
    def data(self) -> T:
        return self._data
    
    
    @property
    def invokation(self) -> "Invokation":
        return self._invokation_ref()
    

    @property
    def chain(self) -> list["Invokation"]:
        return self._invokation_ref().chain
    
    

    def add_next(self, invokation: "Invokation"):
        self.invokation.add_next(invokation)


    def processed(self):
        self.invokation.one_output_processed()


class Invokation[D]:
    __slots__ = (
        'id', 
        '_route', 
        'outputs', 
        'inside', 
        'prev', 
        'nexts',
        '_event',
        '_marked_completed', 
        '_on_output_callbacks', 
        '_on_next_callbacks', 
        '_on_complete_callbacks', 
        '_on_chain_dead_callbacks', 
        '_unprocessed_output_count_ref', 
        '_alive_chain_count_ref', 
        '_future',
        '__weakref__'
    )

    def __init__(self, id=None):
        self.id = id or uuid.uuid4().hex
        self._route = None
        self.outputs: list[Output[D]] | None = None
        self.inside: Invokation | None = None
        self.prev: ReferenceType["Invokation"] | None = None
        self.nexts: list["Invokation"] | None = None
        self._event = asyncio.Condition()
        self._marked_completed = False
        self._on_output_callbacks: list[Callable[[Output], None]] = []
        self._on_next_callbacks: list[Callable[["Invokation"], None]] = []
        self._on_complete_callbacks: list[Callable[[list[Output]], None]] = []
        self._on_chain_dead_callbacks: list[Callable[["Output"], None]] = []
        self._unprocessed_output_count_ref: AliveCountRef = AliveCountRef()
        self._alive_chain_count_ref: AliveCountRef = AliveCountRef(1)

        def if_chain_dead():
            if not self.nexts: 
                for callback in self._on_chain_dead_callbacks: callback()
        self._unprocessed_output_count_ref.on_zero_alive(lambda: if_chain_dead())
        self._unprocessed_output_count_ref.on_zero_alive(lambda: self._alive_chain_count_ref.decrement())

        self._future: asyncio.Future[list[Output[D]]] = asyncio.Future()


    @classmethod
    def wraps(cls, inside: "Invokation", *args, **kwargs):
        self = cls(*args, **kwargs)
        self.inside = inside
        self.inside.on_chain_dead(lambda output: self.add_output(output.data))
        self.inside.on_all_chains_dead(lambda: self.mark_completed())
        return self


    @property
    def route(self) -> str:
        return self._route


    @property
    def chain(self) -> list["Invokation"]:
        if self.prev:
            prev = self.prev()
            if prev is None: 
                raise ValueError("Potential bug! Previous got referenced. Unable to form chain")
            return [*prev.chain, self]
        return [self]

    
    @property
    def unprocessed_outputs_count(self) -> int:
        return self._unprocessed_output_count_ref.count

    @property
    def is_completed(self) -> bool:
        return self._marked_completed
       

    @property
    def is_dead(self) -> bool:
        return self._unprocessed_output_count_ref.dropped_to_zero
    

    @property
    def chain_count(self) -> int:
        if self.nexts:
            total = 0
            for next in self.nexts:
                total += next.chain_count
            return total
        return 1

    
    @property
    def is_end_of_chain(self) -> bool:
        return self.is_dead and not self.nexts
    

    @property
    def is_chain_dead(self) -> bool:
        return self._alive_chain_count_ref.dropped_to_zero

    
    def __matmul__(self, other: str) -> "Invokation":
        if self._route:
            self._route = other.strip('/') + '/' + self._route
        else:
            self._route = other.strip('/') + '/'
        if self.inside: self.inside @ other
        return self
        

    def __await__(self):
        return self._future.__await__()
    

    def __aiter__(self):
        return self._iter_outputs()
    
    
    async def _iter_outputs(self):
        idx = 0
        while True:
            # fast-path
            if self.outputs and idx < len(self.outputs):
                item = self.outputs[idx]
                idx += 1
                yield item
                continue

            if self.is_completed:
                if not self.outputs:
                    raise ValueError("Invokation marked complete without any output. Likely, calling the agent is not invoking it.")
                return

            async with self._event:
                # double-check the condition under the lock to avoid missed wakeups
                should_wait = (not self.is_completed) and (not self.outputs or idx >= len(self.outputs))
                if should_wait:
                    await self._event.wait()


    def add_output(self, output: D):
        if self.is_completed:
            raise ValueError("Cannot add new output after marking this invokation complete")
        o = Output(output, self)
        if self.outputs:
            self.outputs.append(o)
        else:
            self.outputs = [o]
        self._unprocessed_output_count_ref.increment()
        async def _notify():
            async with self._event:
                self._event.notify_all()
        asyncio.create_task(_notify())
        for callback in self._on_output_callbacks: callback(o)


    
    def add_next(self, next: "Invokation"):
        if self.is_dead:
            raise ValueError("Cannot chain new invokation after marking this output dead")
        next.register_on_chain_dead_callbacks(self._on_chain_dead_callbacks)
        if self.nexts: self.nexts.append(next)
        else: self.nexts = [next]
        self._alive_chain_count_ref.increment()
        next.on_all_chains_dead(lambda _: self._alive_chain_count_ref.decrement())
        next.prev = ref(self)
        for callback in self._on_next_callbacks: callback(next)

    
    def one_output_processed(self):
        self._unprocessed_output_count_ref.decrement()


    def mark_completed(self):
        if self.is_completed:
            raise ValueError("Invokation already marked completed once")
        self._future.set_result(self.outputs)
        self._marked_completed = True
        async def _notify():
            async with self._event:
                self._event.notify_all()
        asyncio.create_task(_notify())
        for callback in self._on_complete_callbacks: callback(self.outputs)
        if not self.outputs: self._unprocessed_output_count_ref.zero()


    def on_output(self, func: Callable[[Output[D]], None]):
        self._on_output_callbacks.append(func)


    def on_complete(self, func: Callable[[list[Output[D]]], None]):
        self._on_complete_callbacks.append(func)


    def on_next(self, func: Callable[["Invokation"], None]):
        self._on_next_callbacks.append(func)


    def on_dead(self, func: Callable[[], None]):
        self._unprocessed_output_count_ref.on_zero_alive(func)


    def on_chain_dead(self, func: Callable[["Output"], None]):
        self._on_chain_dead_callbacks.append(func)
        if self.nexts: 
            for next in self.nexts: next.on_chain_dead(func)


    def on_all_chains_dead(self, func: Callable[[Output], None]):
        self._alive_chain_count_ref.on_zero_alive(func)


    def register_on_next_callbacks(self, funcs: list[Callable[["Invokation"], None]]):
        self._on_next_callbacks.extend(funcs)


    def register_on_chain_dead_callbacks(self, funcs: list[Callable[[Output], None]]):
        self._on_chain_dead_callbacks.extend(funcs)
        if self.nexts: 
            for next in self.nexts: next.register_on_chain_dead_callbacks(funcs)
    

    def find_by_id(self, id: str) -> Invokation | None:
        if self.id == id:
            return self
        for next in self.nexts:
            found = next.find_by_id(id)
            if found:
                return found
        raise ValueError(f"InvokationChain with id {id} not found in chain starting with {self.id}")
    

    def copy_without_outputs(self):
        return Invokation(id=self.id)

    
    def normalise(self) -> list[list["Invokation"]]:
        chains = []
        if self.nexts:
            for next in self.nexts:
                chains.extend(next.normalise())
            return chains
        else:
            return [self.chain]


import importlib
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from vespwood_openai import OpenAIChatCompletionGenerator, OpenAIResponsesGenerator
    
def __getattr__(name):
    try:
        # We dynamically load the partner package only when needed
        module = importlib.import_module("vespwood_openai")
        return getattr(module, name)
    except ImportError:
        raise ImportError(
            "The 'vespwood-openai' package is required for this feature. "
            "Please install it using: pip install vesp[openai]"
        )


from abc import abstractmethod
from typing import Callable, Any

from vesp.visibility import Visibility
from vesp.invokation import Invokation
from .agent import BaseAgent
from .team import AgentsTeam, HandoverResponse
from vespwood import Completor, GeneratorClass, Generator


class GoalSeakerTeam(AgentsTeam):
    def __init__(self, generator: Callable[..., Completor], *args, **kwargs):
        self.__generator__: Completor = generator(*args, **kwargs)
        super().__init__(self.create_routes(), *args, **kwargs)

    @property
    def goal(self) -> str:
        return self.__doc__
    
    
    @abstractmethod
    def create_channels(self) -> dict[str, list[str]]:
        pass


    def handover(self, route: str, output: Any, chain: list[Invokation]) -> list[HandoverResponse] | None:
        pass



def goal_seaker_team[T: GoalSeakerTeam](cls: type[T] | None = None, /, *, entrypoint: str = "/", generator: GeneratorClass | Generator | None):
    def decorator(cls: type[T]) -> type[T]:
        # Assert class is GoalSeakerTeam subclass
        if not issubclass(cls, GoalSeakerTeam):
            raise TypeError("goal_seaker_team decorator can only be used with subclass of GoalSeakerTeam")
        
        # Actual decorator implementation
        class GoalSetterTeamWrapper(cls):
            def __init__(self, *args, **kwargs):
                self.__name__ = cls.__name__
                self.__doc__ = cls.__doc__
                super().__init__(generator, *args, **kwargs)
                if self.index:
                    self[entrypoint] = self.index - Visibility.PUBLIC


            @property
            def index(self) -> BaseAgent | None:
                return self.get(entrypoint)

        
        return GoalSetterTeamWrapper
    if cls:
        return decorator(cls)
    else:
        return decorator



import asyncio
from pathlib import Path
from typing import AsyncIterator, Callable, Concatenate, Literal, Any
from abc import ABCMeta, abstractmethod
from vespwood import (
    FormatKeys,
    Generator,
    GeneratorClass,
    Interceptor,
    Schema,
    Hook,
    Invokation,
    PreparedArgs, 
    TaggedMessages, 
    Completor,
    Validator,
    Invokable,
    to_json_schema
)
import inspect


class AgentMeta(ABCMeta):
    def __sub__(cls, other: Literal["public", "private"]) -> "AgentMeta":
        class ScopedAgent(cls):
            _accessibility = other

        ScopedAgent.__class__.__name__ = cls.__name__
        ScopedAgent.__class__.__qualname__ = cls.__qualname__
        return ScopedAgent

        

class BaseAgent(Invokable, metaclass=AgentMeta):  
    _accessibility = "private"

    def __sub__(self, other: Literal["public", "private"]) -> "BaseAgent":
        self._accessibility = other
        return self


    @property
    def is_public(self):
        return self._accessibility == "public"


    @abstractmethod
    def __call__(self, *args, **kwargs) -> Invokation:
        pass


class Agent[O](BaseAgent, metaclass=AgentMeta):
    def __init__(self, *args, **kwargs):
        self._name = self.__class__.__name__
        self._description = self.__doc__
        self._schema = to_json_schema(self)
        super().__init__()

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def description(self) -> str:
        return self._description

    @property
    def schema(self) -> dict[str, Any]:
        return self._schema
    

    @abstractmethod
    async def invoke(self, args: PreparedArgs) -> tuple[TaggedMessages, FormatKeys]:
        '''Accepts args_list, and adds outputs to the Invokation chain object'''
        ...


    @abstractmethod
    async def handle_responses(self, messages: TaggedMessages, format_keys: FormatKeys) -> O:
        ...


    def __get_output__(self, messages: TaggedMessages, format_keys: FormatKeys, *, future: asyncio.Future | None = None, chain: Invokation[O] | None = None) -> O:
        def on_output(output: O):
            if chain: chain.add_output(output)
            if future: future.set_result(output)
        # Step 3: Handle Response
        handle_response_task = asyncio.create_task(self.handle_responses(messages, format_keys))
        handle_response_task.add_done_callback(lambda t: on_output(t.result()))
        

    def __call__(self,  args: PreparedArgs) -> Invokation[O]:
        chain = Invokation()
        async def run_with() -> O:
            result = await self.invoke(args)
            return await self.__get_output__(*result, chain=chain)
        task = asyncio.create_task(run_with())
        task.add_done_callback(lambda _: chain.mark_completed())
        return chain
    

def agent[T: Agent](
        cls: type[T] | None = None, /, *,
        generator: GeneratorClass | Generator | None, 
        prompt_structure: str | None = None, 
        max_requests: int = 0, 
        delay_constant: int = 0, 
        schemas: list[Schema] = [],
        validators: list[Validator] = [], 
        tools: list[Invokable] = [], 
        hooks: list[Hook] = []):
    def decorator(cls: type[T]) -> type[T]:
        if not issubclass(cls, Agent):
            raise TypeError("agent decorator can only be used with subclass of Agent")
    
        class AgentWrapper(cls):
            def __init__(self, interceptors: list[Interceptor] = [], *args, **kwargs):
                _generator: Generator = generator(
                    *args,
                    **kwargs
                ) if isinstance(generator, GeneratorClass) else generator
                src_file = inspect.getsourcefile(cls)
                _, src_line = inspect.getsourcelines(cls)
                _prompt_structure = prompt_structure
                if isinstance(prompt_structure, str):
                    # Convert relative path to absolute path
                    path = Path(prompt_structure)
                    if not path.is_absolute() and not path.is_file():
                        path = (Path(src_file).parent / path)
                        _prompt_structure = str(path)
                try:
                    self.__completer = Completor(_generator,
                        prompt_structure=_prompt_structure, 
                        validators=validators, 
                        delay_constant=delay_constant, 
                        max_requests=max_requests, 
                        schemas=schemas,
                        tools=tools,
                        hooks=hooks,
                        interceptors=interceptors
                    )
                except FileNotFoundError as e:
                    e.add_note(f'File "{Path(src_file)}", line {src_line}, in {cls.__qualname__}')
                    raise 

                super().__init__(*args, **kwargs)
                if self.__completer.name != "":
                    self._name = self.__completer.name
                if self.__completer.description is not None and self.__completer.description != "":
                    self.description = self.__completer.description


            async def invoke(self, args: PreparedArgs) -> tuple[TaggedMessages, FormatKeys]:
                return await self.__completer(args)


            def __str__(self):
                return cls.__name__


            def __repr__(self):
                return cls.__name__
            
        AgentWrapper.__name__ = cls.__name__
        AgentWrapper.__qualname__ = cls.__qualname__
        return AgentWrapper
    
    if cls:
        return decorator(cls)
    else:
        return decorator


def returns_args[**I, O](func: Callable[Concatenate[Agent[O], I], PreparedArgs]) -> Callable[Concatenate[Agent[O], I], Invokation[O]]:
    def fn(self: Agent[O], *args: I.args, **kwargs: I.kwargs) -> Invokation[O]:
        chain = Invokation()
        async def run_with():
            completion_futures: list[asyncio.Future] = []
            # Step 1: Prepare Args
            prepared_args_list = await func(self, *args, **kwargs)
            for prepared_args in prepared_args_list:
                #Step 2: Invoke
                future = asyncio.Future()
                completion_futures.append(future)
                invokation_task = asyncio.create_task(self.invoke(prepared_args))
                # Step 3: Handle Response
                invokation_task.add_done_callback(lambda t: self.__get_output__(*t.result(), future=future, chain=chain))
            return await asyncio.gather(*completion_futures)
        task = asyncio.create_task(run_with())
        task.add_done_callback(lambda _: chain.mark_completed())
        return chain
    return fn


def yields_args[**I, O](func: Callable[Concatenate[Agent[O], I], AsyncIterator[PreparedArgs]]) -> Callable[Concatenate[Agent[O], I], Invokation[O]]:
    def fn(self: Agent[O], *args: I.args, **kwargs: I.kwargs) -> Invokation[O]:
        chain = Invokation()
        async def run_with():
            completion_futures: list[asyncio.Future] = []
            # Step 1: Prepare Args
            async for prepared_args in func(self, *args, **kwargs):
                #Step 2: Invoke
                future = asyncio.Future()
                completion_futures.append(future)
                invokation_task = asyncio.create_task(self.invoke(prepared_args))
                # Step 3: Handle Response
                invokation_task.add_done_callback(lambda t: self.__get_output__(*t.result(), future=future, chain=chain))
            return await asyncio.gather(*completion_futures)
        task = asyncio.create_task(run_with())
        task.add_done_callback(lambda _: chain.mark_completed())
        return chain
    return fn


from vespwood import (
    PromptStructure, MessageList,
    Completor, PromptMapping,
    Logic, Expression, match
)

__all__ = [
    "PromptStructure",
    "MessageList",
    "Completor",
    "PromptMapping",
    "Logic",
    "Expression",
    "match"
]


import importlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vespwood_anthropic import AnthropicMessagesGenerator
    
def __getattr__(name):
    try:
        module = importlib.import_module("vespwood_anthropic")
        return getattr(module, name)
    except ImportError:
        raise ImportError(
            "The 'vespwood-anthropic' package is required for this feature. "
            "Please install it using: pip install vesp[anthropic]"
        )


from abc import abstractmethod
import asyncio
from typing import List, Optional, Callable, Dict, Any, Tuple, TypeVar, Type
from vesp.invokation import Invokation, Output
from .agent import BaseAgent
from vesp.visibility import Visibility

type AgentLike = BaseAgent | Callable[..., BaseAgent]
type TeamLike = Dict[str, "TeamLike" | AgentLike]

type Next = str 
type Args = List[Any]
type Kwargs = Dict[str, Any]
type NextWithArgs = Tuple[Next, Args]
type NextWithKwargs = Tuple[Next, Kwargs]
type NextWithArgsAndKwargs = Tuple[Next, Args, Kwargs]

type HandoverResponse = Next | NextWithArgs | NextWithKwargs | NextWithArgsAndKwargs

def normalise(routes: TeamLike, prefix: str = "") -> Dict[str, BaseAgent]:
    normalised_dict = {}
    for key, value in routes.items():
        if isinstance(value, dict):
            normalised_dict.update(normalise(value, f"{prefix}/{key}"))
        else:
            normalised_dict.update({f"{prefix}/{key}": value})
    return normalised_dict


  
class Chain(list[Invokation]):
    def __init__(self, *invokation: List[Invokation]):
        super().__init__(invokation)

    def with_next(self, invokation: Invokation) -> "Chain":
        chain = Chain(self)
        chain.append(invokation)
        return chain


class AgentsTeam(dict, BaseAgent):
    def __init__(self, route_map: Dict[str, any], *args, **kwargs):
        for route, agent_class in route_map.items():
            route = route.removeprefix('/')
            paths = route.split('/')
            if isinstance(agent_class, dict):
                self.update({paths.pop(0): AgentsTeam({'/'.join(paths): agent_class} if len(paths) > 0 else agent_class, *args, **kwargs)})
            elif isinstance(agent_class, BaseAgent):
                self.update({paths.pop(0): agent_class})
            else:
                self.update({paths.pop(0): AgentsTeam({'/'.join(paths): agent_class}, *args, **kwargs) if len(paths) > 0 else agent_class(*args, **kwargs)})
        self.__args = args
        self.__kwargs = kwargs


    def normalise(self) -> Dict[str, BaseAgent]:
        return normalise(self)
            

    @abstractmethod
    def create_routes(self) -> Dict[str, any]:
        pass

    
    @property
    @abstractmethod
    def entrypoint(self) -> str:
        pass
    
    
    @property
    def index(self) -> BaseAgent:
        return self[self.entrypoint]


    def __getitem__(self, key: str) -> BaseAgent:
        if key.__contains__('/'):
            key = key.removeprefix('/')
            paths = key.split('/')
            agent = self
            for path in paths:
                agent = agent[path.strip()]
            if agent is None:
                raise ValueError(f"No agent found at {key} in team {self.name}")
            return agent
        else:
            agent = super().__getitem__(key)
            if agent is None:
                raise ValueError(f"No agent found at {key} in team {self.name}")
            return agent
        

    def __setitem__(self, key: str, value: TeamLike | AgentLike):
        key = key.removeprefix('/')
        paths = key.split('/')
        agent = self
        while len(paths) > 0:
            path = paths.pop(0)
            if len(paths) == 0:
                if isinstance(value, dict):
                    agent.update({path: AgentsTeam(value, *self.__args, **self.__kwargs)})                        
                elif isinstance(value, BaseAgent):
                    agent.update({path: value})
                else:
                    agent.update({path: value(*self.__args, **self.__kwargs)})
                return
            
            if not agent.get(path):
                if isinstance(value, dict):
                    agent.update({path: AgentsTeam({'/'.join(paths): value})})
                elif isinstance(value, BaseAgent):
                    agent.update({path: value})
                else:
                    agent.update({path: AgentsTeam({'/'.join(paths): value}, *self.__args, **self.__kwargs)})
                return

            agent = agent[path]
            


    def __contains__(self, key: str) -> bool:
        if key.__contains__('/'):
            key = key.removeprefix('/')
            paths = key.split('/')
            agent = self
            for path in paths:
                if not agent.get(path):
                    return False
                agent = agent[path]
            return True
        else:
            return super().__contains__(key)
        
    
    def get(self, key: str, default: Optional[BaseAgent] = None):
        if key in self:
            return self[key]
        return default
        
    
    def __add__(self, other: TeamLike) -> "AgentsTeam":
        other = normalise(other)
        for key, value in other.items():
            self[key] = value
        return self
    
        
    def __radd__(self, other: TeamLike) -> "AgentsTeam":
        other = normalise(other)
        for key, value in other.items():
            self[key] = value
        return self
    
    
    def __iadd__(self, other: TeamLike) -> "AgentsTeam":
        other = normalise(other)
        for key, value in other.items():
            self[key] = value
        return self


    @abstractmethod
    async def handover(self, route: str, output: Any, chain: list[Invokation]) -> Optional[List[HandoverResponse]]:
        pass


    async def __handover__(self, route: str, output: Output) -> None:
        handovers = await self.handover(route, output.data, output.chain)
        if handovers:
            for response in handovers:
                next_route, args, kwargs = response
                if isinstance(response, str):
                    next_route = response
                else:
                    if isinstance(response, (str, list)):
                        next_route, args = response
                    elif isinstance(response, (str, dict)):
                        next_route, kwargs = response
                    else:
                        next_route, args, kwargs = response
                agent = self[next_route]
                next: Invokation = agent(*args, **kwargs) @ route
                output.add_next(next)
                next.on_output(lambda o: asyncio.create_task(self.__handover__(next_route, o)))
        output.processed()
                

    def __call__(self, *args, **kwargs) -> Invokation:
        route = kwargs.get('route') or self.entrypoint
        agent = self[route]
        if agent.is_not_public:
            raise ValueError("Agent ", agent, "in team ", self.name, "cannot be invoked publicly")
        invokation = agent(*args, **kwargs) @ route
        invokation.on_output(lambda o: asyncio.create_task(self.__handover__(route, o)))
        wrapper = Invokation.wraps(invokation)
        return wrapper


    @property
    def name(self):
        return self.__name__


    @property
    def description(self) -> str:
        return self.__doc__ or ""


    @property
    def schema(self) -> Dict[str, Any]:
        routes = self.normalise()
        properties: dict[str, any] = self.index.schema["properties"]
        oneOf = [{
            "properties": properties,
            "required": [ *properties ],
            "additionalProperties": False
        }]
        for route, agent in routes.items():
            if agent.is_public:
                oneOf.append({
                    "properties": {
                        "route": { "const": route },
                        **agent.schema["properties"]
                    },
                    "required": ["route", *agent.schema["properties"]],
                    "additionalProperties": False
                })

        return {
             "type": "object",
             "properties": properties,
             "oneOf": oneOf
        }

T = TypeVar('T', bound=AgentsTeam)

def team(entrypoint: str = "/"):
    def decorator(cls: Type[T]):
        # Assert class is AgentsTeam subclass
        if not issubclass(cls, AgentsTeam):
            raise TypeError("team decorator can only be used with subclass of AgentsTeam")
        
        # Actual decorator implementation
        class TeamWrapper(cls):        
            def __init__(self, *args, **kwargs):
                self.__name__ = cls.__name__
                self.__doc__ = cls.__doc__
                super().__init__(self.create_routes(), *args, **kwargs)
                if self.index:
                    self[entrypoint] = self.index - Visibility.PUBLIC


            @property
            def entrypoint(self) -> str:
                return entrypoint

        return TeamWrapper
    return decorator


# Ideas: Other team types 
# - Simple State Machine uses provided list of routes to simply follow the list
